import json
from dcicutils.ff_utils import get_metadata, search_metadata
from dcicutils.creds_utils import CGAPKeyManager


class VariantUtils:

    SEARCH_VARIANTS_BY_GENE = (f'/search/?type=VariantSample&limit=1'
                               f'&variant.genes.genes_most_severe_gene.display_title=')
    SEARCH_RARE_VARIANTS_BY_GENE = (f'/search/?samplegeno.samplegeno_role=proband&type=VariantSample'
                                    f'&variant.csq_gnomadg_af_popmax.from=0&variant.csq_gnomadg_af_popmax.to=0.001'
                                    f'&variant.genes.genes_most_severe_gene.display_title=')

    def __init__(self, *, env_name) -> None:
        self._key_manager = CGAPKeyManager()
        self.creds = self._key_manager.get_keydict_for_env(env=env_name)
        # Uncomment this if needed
        # self.health = get_health_page(key=self.creds)
        self.base_url = self.creds['server']

    def get_creds(self):
        return self.creds

    def get_rare_variants_by_gene(self, *, gene, sort, addon=''):
        """Searches for rare variants on a particular gene"""
        return search_metadata(f'{self.base_url}/{self.SEARCH_RARE_VARIANTS_BY_GENE}{gene}\
                               &sort=-{sort}{addon}', key=self.creds)

    def find_number_of_sample_ids(self, gene):
        """Returns the number of samples that have a mutation on the specified gene"""
        return len(set(variant.get('CALL_INFO')
                       for variant in self.get_rare_variants_by_gene(gene=gene, sort='variant.ID')))

    def get_total_result_count_from_search(self, gene):
        """Returns total number of variants associated with specified gene"""
        res = get_metadata(self.SEARCH_VARIANTS_BY_GENE + gene, key=self.creds)
        return res['total']

    @staticmethod
    def sort_dict_in_descending_order(unsorted_dict):
        """Sorts dictionary in descending value order"""
        sorted_list = sorted(unsorted_dict.items(), key=lambda x: x[1], reverse=True)
        return dict(sorted_list)

    def create_dict_of_mutations(self, gene):
        """Creates dictionary of specified gene and mutations that occur 10+ times in database, in the form:
            {gene: {mutation1 pos: #variants, mutation2 pos: #variants, ...}"""
        mutation_dict = {}
        unique_positions = set()
        for variant in self.get_rare_variants_by_gene(gene=gene, sort='variant.ID'):
            pos = variant['variant']['POS']
            if pos not in unique_positions:
                unique_positions.add(pos)
                mutation_dict[pos] = 1
            else:
                mutation_dict[pos] += 1
        return {gene: self.sort_dict_in_descending_order({k: v for k, v in mutation_dict.items() if v >= 10})}

    @staticmethod
    def return_json(file_name):
        with open(file_name, 'r') as f:
            file_content = json.loads(f)
        return file_content

    @staticmethod
    def create_dict_from_json_file(file_name):
        """Creates dictionary object from specified json file"""
        with open(file_name) as f:
            json_list = f.read()
        return json.loads(json_list)

    def create_list_of_msa_genes(self):
        """Creates list of genes relating to the brain or nervous system
            (determined by whether keywords 'neur' or 'nerv' in summary)"""
        genes = self.return_json('gene.json')
        return [gene['gene_symbol'] for gene in genes
                if 'nerv' in gene.get('gene_summary', '')
                or 'neur' in gene.get('gene_summary', '')]

    def create_url(self, gene):
        """Returns a url to the variants at the most commonly mutated position of specified gene"""
        d = self.create_dict_from_json_file('10+sorted_msa_genes_and_mutations.json')
        pos = list(d[gene].keys())[0]
        return self.SEARCH_RARE_VARIANTS_BY_GENE + gene + f'&variant.POS.from={pos}&variant.POS.to={pos}&sort=-DP'

    def create_list_of_als_park_genes(self):
        """Creates list of genes that relating to Parkinson's or ALS
            (determined by whether keywords 'Parkinson' or 'ALS' in summary)"""
        genes = self.return_json('gene.json')
        return [gene['gene_symbol'] for gene in genes
                if 'Parkinson' in gene.get('gene_summary', '')
                or 'ALS' in gene.get('gene_summary', '')]



import os
from typing import Any, Dict, List, Optional, Tuple
from dcicutils.misc_utils import to_camel_case


class JsonSchemaConstants:
    ANY_OF = "anyOf"
    ARRAY = "array"
    BOOLEAN = "boolean"
    DEFAULT = "default"
    ENUM = "enum"
    FORMAT = "format"
    INTEGER = "integer"
    ITEMS = "items"
    NUMBER = "number"
    OBJECT = "object"
    ONE_OF = "oneOf"
    PATTERN = "pattern"
    PROPERTIES = "properties"
    REF = "$ref"
    REQUIRED = "required"
    STRING = "string"
    TYPE = "type"


class EncodedSchemaConstants:
    IDENTIFYING_PROPERTIES = "identifyingProperties"
    LINK_TO = "linkTo"
    MERGE_REF = "$merge"
    MIXIN_PROPERTIES = "mixinProperties"
    UNIQUE_KEY = "uniqueKey"


class SchemaConstants(JsonSchemaConstants, EncodedSchemaConstants):
    pass


def get_properties(schema: Dict[str, Any]) -> Dict[str, Any]:
    """Return the properties of a schema."""
    return schema.get(SchemaConstants.PROPERTIES, {})


def get_property(schema: Dict[str, Any], property_name: str) -> Dict[str, Any]:
    """Return property in properties, if found."""
    return get_properties(schema).get(property_name, {})


def has_property(schema: Dict[str, Any], property_name: str) -> bool:
    """Return True if the schema has the given property."""
    return property_name in get_properties(schema)


def get_required(schema: Dict[str, Any]) -> List[str]:
    """Return the required properties of a schema."""
    return schema.get(SchemaConstants.REQUIRED, [])


def get_pattern(schema: Dict[str, Any]) -> str:
    """Return the pattern property of a schema."""
    return schema.get(SchemaConstants.PATTERN, "")


def get_any_of(schema: Dict[str, Any]) -> List[Dict[str, Any]]:
    """Return the anyOf properties of a schema."""
    return schema.get(SchemaConstants.ANY_OF, [])


def get_one_of(schema: Dict[str, Any]) -> List[Dict[str, Any]]:
    """Return the oneOf properties of a schema."""
    return schema.get(SchemaConstants.ONE_OF, [])


def get_conditional_required(schema: Dict[str, Any]) -> List[str]:
    """Get required + possibly required properties.

    Using heuristics here; update as needed.
    """
    return sorted(
        list(
            set(
                get_required(schema)
                + get_any_of_required(schema)
                + get_one_of_required(schema)
            )
        )
    )


def get_any_of_required(schema: Dict[str, Any]) -> List[str]:
    """Get required properties from anyOf."""
    return [
        property_name
        for any_of_schema in get_any_of(schema)
        for property_name in get_required(any_of_schema)
    ]


def get_one_of_required(schema: Dict[str, Any]) -> List[str]:
    """Get required properties from oneOf."""
    return [
        property_name
        for one_of_schema in get_one_of(schema)
        for property_name in get_required(one_of_schema)
    ]


def get_mixin_properties(schema: Dict[str, Any]) -> List[Dict[str, Any]]:
    """Return the mixin properties of a schema."""
    return schema.get(SchemaConstants.MIXIN_PROPERTIES, [])


def get_identifying_properties(schema: Dict[str, Any]) -> List[str]:
    """Return the identifying properties of a schema."""
    return schema.get(SchemaConstants.IDENTIFYING_PROPERTIES, [])


def get_schema_type(schema: Dict[str, Any]) -> str:
    """Return the type of a schema."""
    return schema.get(SchemaConstants.TYPE, "")


def is_array_schema(schema: Dict[str, Any]) -> bool:
    """Return True if the schema is an array."""
    return get_schema_type(schema) == SchemaConstants.ARRAY


def is_object_schema(schema: Dict[str, Any]) -> bool:
    """Return True if the schema is an object."""
    return get_schema_type(schema) == SchemaConstants.OBJECT


def is_string_schema(schema: Dict[str, Any]) -> bool:
    """Return True if the schema is a string."""
    return get_schema_type(schema) == SchemaConstants.STRING


def is_number_schema(schema: Dict[str, Any]) -> bool:
    """Return True if the schema is a number."""
    return get_schema_type(schema) == SchemaConstants.NUMBER


def is_integer_schema(schema: Dict[str, Any]) -> bool:
    """Return True if the schema is an integer."""
    return get_schema_type(schema) == SchemaConstants.INTEGER


def is_boolean_schema(schema: Dict[str, Any]) -> bool:
    """Return True if the schema is a boolean."""
    return get_schema_type(schema) == SchemaConstants.BOOLEAN


def get_items(schema: Dict[str, Any]) -> Dict[str, Any]:
    """Return the items of a schema."""
    return schema.get(SchemaConstants.ITEMS, {})


def get_format(schema: Dict[str, Any]) -> str:
    """Return the format of a schema."""
    return schema.get(SchemaConstants.FORMAT, "")


def get_conditional_formats(schema: Dict[str, Any]) -> List[str]:
    """Return the format of a schema, as directly given or as listed
    as an option via oneOf or anyOf.
    """
    formats = set(
        get_any_of_formats(schema) + get_one_of_formats(schema) + [get_format(schema)]
    )
    return sorted(list([format_ for format_ in formats if format_]))


def get_any_of_formats(schema: Dict[str, Any]) -> List[str]:
    """Return the formats of a schema's anyOf properties."""
    return [
        get_format(any_of_schema)
        for any_of_schema in get_any_of(schema)
        if get_format(any_of_schema)
    ]


def get_one_of_formats(schema: Dict[str, Any]) -> List[str]:
    """Return the formats of a schema's oneOf properties."""
    return [
        get_format(one_of_schema)
        for one_of_schema in get_one_of(schema)
        if get_format(one_of_schema)
    ]


class Schema:

    def __init__(self, schema: dict, type: Optional[str] = None) -> None:
        self._data = schema if isinstance(schema, dict) else (schema.data if isinstance(schema, Schema) else {})
        self._type = ((type if isinstance(type, str) else "") or
                      Schema.type_name(self._data.get("title", "")) or
                      Schema.type_name(self._data.get("$id", "")))

    @property
    def data(self) -> dict:
        return self._data

    @property
    def type(self) -> str:
        return self._type

    @staticmethod
    def type_name(value: str) -> Optional[str]:  # File or other name.
        if isinstance(value, str) and (value := os.path.basename(value.replace(" ", ""))):
            return to_camel_case(value[0:dot] if (dot := value.rfind(".")) >= 0 else value)

    def property_by_path(self, property_path: str) -> Optional[dict]:
        """
        Looks for the given property path within this Portal schema and returns its dictionary value.
        This given property path can be either a simple property name, or a series of dot-separated
        property names representing nested (object) properties; and/or the property names may also
        be suffixed with a pound (#) characteter, optionally followed by an integer, representing
        an array type property and its optional array index (this integer, if specified, is ignored
        for the purposes of this function, but it may have been created by another process/function,
        for example by PortalObject.compare). If the property is not found then None is returned.
        """
        return Schema.get_property_by_path(self._data, property_path)

    _ARRAY_NAME_SUFFIX_CHAR = "#"
    _DOTTED_NAME_DELIMITER_CHAR = "."

    @staticmethod
    def get_property_by_path(schema: dict, property_path: str) -> Optional[dict]:
        if not isinstance(schema, dict) or not isinstance(property_path, str):
            return None
        elif not (schema_properties := schema.get(JsonSchemaConstants.PROPERTIES)):
            return None
        property_paths = property_path.split(Schema._DOTTED_NAME_DELIMITER_CHAR)
        for property_index, property_name in enumerate(property_paths):
            property_name, array_specifiers = Schema._unarrayize_property_name(property_name)
            if not (property_value := schema_properties.get(property_name)):
                return None
            elif (property_type := property_value.get(JsonSchemaConstants.TYPE)) == JsonSchemaConstants.OBJECT:
                property_paths_tail = Schema._DOTTED_NAME_DELIMITER_CHAR.join(property_paths[property_index + 1:])
                return Schema.get_property_by_path(property_value, property_paths_tail)
            elif (property_type := property_value.get(JsonSchemaConstants.TYPE)) == JsonSchemaConstants.ARRAY:
                if not array_specifiers:
                    if property_index == len(property_paths) - 1:
                        return property_value
                    return None
                for array_index in range(len(array_specifiers)):
                    if property_type != JsonSchemaConstants.ARRAY:
                        return None
                    elif not (array_items := property_value.get(JsonSchemaConstants.ITEMS)):
                        return None
                    property_type = (property_value := array_items).get(JsonSchemaConstants.TYPE)
                if property_type == JsonSchemaConstants.OBJECT:
                    if property_index == len(property_paths) - 1:
                        return property_value
                    property_paths_tail = Schema._DOTTED_NAME_DELIMITER_CHAR.join(property_paths[property_index + 1:])
                    return Schema.get_property_by_path(property_value, property_paths_tail)
        return property_value

    @staticmethod
    def _unarrayize_property_name(property_name: str) -> Tuple[str, Optional[List[int]]]:
        if len(components := (property_name := property_name.strip()).split(Schema._ARRAY_NAME_SUFFIX_CHAR)) < 2:
            return property_name, None
        unarrayized_property_name = components[0].strip()
        array_specifiers = []
        for component in components[1:]:
            if component.isdigit():
                array_specifiers.append(int(component))
            elif component == "":
                array_specifiers.append(0)
            else:
                return property_name, None
        return unarrayized_property_name, array_specifiers



import abc
import csv
import openpyxl
from typing import Any, Generator, Iterator, List, Optional, Type, Tuple, Union
from dcicutils.misc_utils import create_dict, right_trim

# Forward type references for type hints.
Excel = Type["Excel"]

# Cell values(s) indicating property deletion.
_CELL_DELETION_VALUES = ["*delete*"]


# Special cell deletion sentinel value (note make sure on deepcopy it remains the same).
class _CellDeletionSentinal(str):
    def __new__(cls):
        return super(_CellDeletionSentinal, cls).__new__(cls, _CELL_DELETION_VALUES[0])
    def __deepcopy__(self, memo):  # noqa
        return self


class RowReader(abc.ABC):

    CELL_DELETION_SENTINEL = _CellDeletionSentinal()

    def __init__(self):
        self.header = None
        self.row_number = 0
        self._warning_empty_headers = False
        self._warning_extra_values = []  # Line numbers.
        self.open()

    def __iter__(self) -> Iterator:
        for row in self.rows:
            self.row_number += 1
            if self.is_comment_row(row):
                continue
            if self.is_terminating_row(row):
                break
            if len(self.header) < len(row):  # Row values beyond what there are headers for are ignored.
                self._warning_extra_values.append(self.row_number)
            yield {column: self.cell_value(value) for column, value in zip(self.header, row)}

    def _define_header(self, header: List[Optional[Any]]) -> None:
        self.header = []
        for index, column in enumerate(header or []):
            if not (column := str(column).strip() if column is not None else ""):
                self._warning_empty_headers = True
                break  # Empty header column signals end of header.
            self.header.append(column)

    def rows(self) -> Generator[Union[List[Optional[Any]], Tuple[Optional[Any], ...]], None, None]:
        yield

    def is_comment_row(self, row: Union[List[Optional[Any]], Tuple[Optional[Any]]]) -> bool:
        return False

    def is_terminating_row(self, row: Union[List[Optional[Any]], Tuple[Optional[Any]]]) -> bool:
        return False

    def cell_value(self, value: Optional[Any]) -> str:
        if value is None:
            return ""
        elif (value := str(value).strip()) in _CELL_DELETION_VALUES:
            return RowReader.CELL_DELETION_SENTINEL
        else:
            return value

    def open(self) -> None:
        pass

    @property
    def file(self) -> Optional[str]:
        return self._file if hasattr(self, "_file") else None

    @property
    def warnings(self) -> List[str]:
        warnings = []
        if self._warning_empty_headers:
            if hasattr(self, "sheet_name") and self.sheet_name:
                src = {"sheet": self.sheet_name}
            else:
                src = {"file": self.file}
            warnings.append({"src": src,
                             "warning": "Empty header column encountered; ignoring it and all subsequent columns."})
        if self._warning_extra_values:
            for row_number in self._warning_extra_values:
                warnings.append({"src": create_dict(file=self.file, row=row_number),
                                 "warning": f"Extra row column values."})
        return warnings


class ListReader(RowReader):

    def __init__(self, rows: List[List[Optional[Any]]]) -> None:
        self._rows = rows
        super().__init__()

    @property
    def rows(self) -> Generator[List[Optional[Any]], None, None]:
        for row in self._rows[1:]:
            yield row

    def open(self) -> None:
        if not self.header:
            self._define_header(self._rows[0] if self._rows else [])


class CsvReader(RowReader):

    def __init__(self, file: str) -> None:
        self._file = file
        self._file_handle = None
        self._rows = None
        super().__init__()

    @property
    def rows(self) -> Generator[List[Optional[Any]], None, None]:
        for row in self._rows:
            yield right_trim(row)

    def open(self) -> None:
        if self._file_handle is None:
            self._file_handle = open(self._file)
            self._rows = csv.reader(self._file_handle, delimiter="\t" if self._file.endswith(".tsv") else ",")
            self._define_header(right_trim(next(self._rows, [])))

    def __del__(self) -> None:
        if (file_handle := self._file_handle) is not None:
            self._file_handle = None
            file_handle.close()


class ExcelSheetReader(RowReader):

    def __init__(self, excel: Excel, sheet_name: str, workbook: openpyxl.workbook.workbook.Workbook) -> None:
        self.sheet_name = sheet_name or "Sheet1"
        self._workbook = workbook
        self._file = excel._file
        self._rows = None
        super().__init__()

    @property
    def rows(self) -> Generator[Tuple[Optional[Any], ...], None, None]:
        for row in self._rows(min_row=2, values_only=True):
            yield right_trim(row)

    def is_terminating_row(self, row: Tuple[Optional[Any]]) -> bool:
        return all(cell is None for cell in row)  # Empty row signals end of data.

    def open(self) -> None:
        if not self._rows:
            self._rows = self._workbook[self.sheet_name].iter_rows
            self._define_header(right_trim(next(self._rows(min_row=1, max_row=1, values_only=True), [])))


class Excel:

    def __init__(self, file: str, reader_class: Optional[Type] = None, include_hidden_sheets: bool = False) -> None:
        self._file = file
        self._workbook = None
        self._include_hidden_sheets = include_hidden_sheets
        self.sheet_names = None
        if isinstance(reader_class, Type) and issubclass(reader_class, ExcelSheetReader):
            self._reader_class = reader_class
        else:
            self._reader_class = ExcelSheetReader
        self.open()

    def sheet_reader(self, sheet_name: str) -> ExcelSheetReader:
        return self._reader_class(self, sheet_name=sheet_name, workbook=self._workbook)

    def open(self) -> None:
        if self._workbook is None:
            import warnings
            with warnings.catch_warnings():
                # Without this warning suppression thing, for some spreadsheets we get this stdout warning:
                # UserWarning: data validation extension is not supported and will be removed
                warnings.filterwarnings("ignore", category=UserWarning)
                self._workbook = openpyxl.load_workbook(self._file, data_only=True)
            self.sheet_names = [sheet_name for sheet_name in self._workbook.sheetnames
                                if not self.is_hidden_sheet(self._workbook[sheet_name])]

    def is_hidden_sheet(self, sheet: openpyxl.worksheet.worksheet.Worksheet) -> bool:
        if not self._include_hidden_sheets:
            if sheet.sheet_state == "hidden":
                return True
            if ((sheet.title.startswith("(") and sheet.title.endswith(")")) or
                (sheet.title.startswith("[") and sheet.title.endswith("]")) or
                (sheet.title.startswith("{") and sheet.title.endswith("}")) or
                (sheet.title.startswith("<") and sheet.title.endswith(">"))):  # noqa
                return True
        return False

    def __del__(self) -> None:
        if (workbook := self._workbook) is not None:
            self._workbook = None
            workbook.close()



from collections import namedtuple
import re
from signal import signal, SIGINT
import sys
import threading
import time
from tqdm import tqdm
from types import FrameType as frame
from typing import Callable, List, Optional, Union
from contextlib import contextmanager
from dcicutils.command_utils import yes_or_no
from dcicutils.misc_utils import find_nth_from_end, format_size, set_nth


class TQDM(tqdm):

    """
    def moveto(self, n):
      # Hack/workaround for apparent tqdm bug where (for example) if we use this twice
      # in a row, and we interrupt (CTRL-C) the first one and abort the task, then the
      # output for the second usage gets a bit garble; on the wrong line and whatnot;
      # somehow, state is stuck across usages; can't quite see how from the tqdm code.
      # This is a bit worrying but so far no other deleterious effects observed.
      # This looks maybe promising:
    return
    """

    # Nevermind the above; found a more pointed solution from here:
    # https://stackoverflow.com/questions/41707229/why-is-tqdm-printing-to-a-newline-instead-of-updating-the-same-line
    # Why in the world would tqdm be maintaining state across instances?? Whatever, this fixes it.
    def __init__(self, *args, **kwargs):
        super()._instances.clear() if super()._instances else None
        super().__init__(*args, **kwargs)


# Wrapper around tqdm command-line progress bar.
class ProgressBar:

    @staticmethod
    @contextmanager
    def define(*args, **kwargs):
        progress_bar = None
        try:
            progress_bar = ProgressBar(*args, **kwargs)
            yield progress_bar
        finally:
            if progress_bar:
                progress_bar.done()

    def __init__(self, total: Optional[int] = None,
                 description: Optional[str] = None,
                 use_byte_size_for_rate: bool = False,
                 catch_interrupt: bool = True,
                 interrupt: Optional[Callable] = None,
                 interrupt_continue: Optional[Callable] = None,
                 interrupt_stop: Optional[Callable] = None,
                 interrupt_exit: bool = False,
                 interrupt_exit_message: Optional[Union[Callable, str]] = None,
                 interrupt_message: Optional[str] = None,
                 tidy_output_hack: bool = True,
                 capture_output_for_testing: bool = False) -> None:
        self._bar = None
        self._started = 0
        self._disabled = False
        self._done = False
        self._tidy_output_hack = (tidy_output_hack is True)
        self._stop_requested = False
        self._use_byte_size_for_rate = (use_byte_size_for_rate is True and self._tidy_output_hack)
        # Interrupt handling. We do not do the actual (signal) interrupt setup
        # in self._initialize as that could be called from a (sub) thread; and in
        # Python we can only set a signal (SIGINT in our case) on the main thread.
        self._catch_interrupt = (catch_interrupt is True)
        self._interrupt = interrupt if callable(interrupt) else None
        self._interrupt_continue = interrupt_continue if callable(interrupt_continue) else None
        self._interrupt_stop = interrupt_stop if callable(interrupt_stop) else None
        if interrupt_exit in [True, False]:
            if not self._interrupt_stop:
                self._interrupt_stop = lambda _: interrupt_exit
            self._interrupt_exit = interrupt_exit
        else:
            self._interrupt_exit = None
        self._interrupt_message = interrupt_message if isinstance(interrupt_message, str) else None
        if isinstance(interrupt_exit_message, str):
            self._interrupt_exit_message = lambda bar: interrupt_exit_message
        elif isinstance(interrupt_exit_message, Callable):
            self._interrupt_exit_message = interrupt_exit_message
        else:
            self._interrupt_exit_message = None
        self._interrupt_handler = None
        if self._catch_interrupt:
            self._interrupt_handler = self._define_interrupt_handler()
        if self._tidy_output_hack is True:
            self._tidy_output_hack = self._define_tidy_output_hack()
        self._total = total if isinstance(total, int) and total >= 0 else 0
        self._description = self._format_description(description)
        self._captured_output_for_testing = [] if capture_output_for_testing else None

    def _initialize(self) -> bool:
        # Do not actually create the tqdm object unless/until we have a positive total.
        if (self._bar is None) and (self._total > 0):
            if self._use_byte_size_for_rate:
                bar_format = "{l_bar}{bar}| {n_fmt}/{total_fmt} | [rate] | {elapsed}{postfix} | ETA: {remaining} "
            else:
                bar_format = "{l_bar}{bar}| {n_fmt}/{total_fmt} | {rate_fmt} | {elapsed}{postfix} | ETA: {remaining} "
            self._bar = TQDM(total=self._total, desc=self._description,
                             dynamic_ncols=True, bar_format=bar_format, unit="", file=sys.stdout)
            self._started = time.time()
            if self._disabled:
                self._bar.disable = True
            return True
        return False

    def set_total(self, value: int, _norefresh: bool = False) -> None:
        if value == self._total:
            # If the total has not changed since last set then do nothing.
            return
        if isinstance(value, int) and value > 0:
            self._total = value
            if self._bar is not None:
                # This reset is needed to get the ETA to reset properly when we reset
                # the total during the course of a single ProgressBar instance.
                self._bar.reset()
                self._bar.total = value
                if not _norefresh:
                    self._bar.refresh()

    def set_progress(self, value: int, _norefresh: bool = False) -> None:
        if isinstance(value, int) and value >= 0:
            if (self._bar is not None) or self._initialize():
                self._bar.n = value
                if not _norefresh:
                    self._bar.refresh()

    def increment_progress(self, value: int) -> None:
        if isinstance(value, int) and value > 0:
            if (self._bar is not None) or self._initialize():
                self._bar.update(value)
                self._bar.refresh()

    def set_description(self, value: str) -> None:
        if isinstance(value, str):
            self._description = self._format_description(value)
            if self._bar is not None:
                # FYI: tqdm.set_description seems to imply a refresh.
                self._bar.set_description(self._description)

    def reset_eta(self) -> None:
        # Since set_total does nothing if total is the same, provide
        # a way to reset the ETA if starting over with the same total.
        # But NOTE that resetting ETA will ALSO reset the ELAPSED time.
        if self._bar is not None:
            progress = self._bar.n
            self._bar.reset()
            self._bar.total = self._total
            self._bar.n = progress
            self._bar.refresh()

    def reset(self, total: int, progress: int = 0, description: Optional[str] = None) -> None:
        self.set_total(total, _norefresh=True)
        self.set_progress(progress, _norefresh=True)
        self.set_description(description)
        self._started = time.time()

    def done(self, description: Optional[str] = None) -> None:
        if self._done or self._bar is None:
            return
        self._ended = time.time()
        self.set_progress(self.total)
        self.set_description(description)
        self._bar.refresh()
        # FYI: Do NOT do a bar.disable = True before a bar.close() or it messes up output
        # on multiple calls; found out the hard way; a couple hours will never get back :-/
        self._bar.close()
        if self._tidy_output_hack:
            self._tidy_output_hack.restore()
        if self._interrupt_handler:
            self._interrupt_handler.restore()
        self._done = True

    def disable(self, value: bool = True) -> None:
        self._disabled = (value is True)
        if self._bar is not None:
            self._bar.disable = self._disabled

    def enable(self, value: bool = True) -> None:
        self.disable(not value)

    @property
    def total(self) -> int:
        return self._bar.total if self._bar else 0

    @property
    def progress(self) -> int:
        return self._bar.n if self._bar else 0

    @property
    def disabled(self) -> bool:
        return self._disabled

    @property
    def enabled(self) -> bool:
        return not self.disabled

    @property
    def stop_requested(self) -> bool:
        return self._stop_requested

    @property
    def captured_output_for_testing(self) -> Optional[List[str]]:
        return self._captured_output_for_testing

    @staticmethod
    def format_captured_output_for_testing(description: str, total: int, progress: int) -> str:
        percent = round((progress / total) * 100.0)
        separator = "✓" if percent == 100 else "|"
        return f"{description} {separator} {percent:>3}% ◀|### | {progress}/{total} | 0.0/s | 00:00 | ETA: 00:00"

    def _format_description(self, value: str) -> str:
        if not isinstance(value, str):
            value = ""
        if self._tidy_output_hack and not value.endswith(self._tidy_output_hack.sentinel):
            value += self._tidy_output_hack.sentinel
        return value

    def _define_interrupt_handler(self) -> None:
        def handle_interrupt(signum: int, frame: frame) -> None:  # noqa
            nonlocal self
            def handle_secondary_interrupt(signum: int, frame: frame) -> None:  # noqa
                print("\nEnter 'yes' or 'no' or CTRL-\\ to completely abort ...")
            self.disable()
            self._interrupt(self) if self._interrupt else None
            set_interrupt_handler(handle_secondary_interrupt)
            if yes_or_no(f"\nALERT! You have interrupted this {self._interrupt_message or 'process'}."
                         f" Do you want to stop{' (exit)' if self._interrupt_exit else ''}?"):
                # Here there was an interrupt (CTRL-C) and the user confirmed (yes)
                # that they want to stop the process; if the interrupt_stop handler
                # is defined and returns True, then we exit the entire process here,
                # rather than simply returning, which is the default.
                if self._interrupt_stop:
                    interrupt_stop = self._interrupt_stop(self)
                    if (interrupt_stop is True) or ((interrupt_stop is None) and (self._interrupt_exit is True)):
                        self.done()
                        restore_interrupt_handler()
                        if self._interrupt_exit_message:
                            if isinstance(interrupt_exit_message := self._interrupt_exit_message(self), str):
                                print(interrupt_exit_message)
                        exit(1)
                    elif interrupt_stop is False or ((interrupt_stop is None) and (self._interrupt_exit is False)):
                        set_interrupt_handler(handle_interrupt)
                        interrupt_continue = self._interrupt_continue(self) if self._interrupt_continue else None
                        if not (interrupt_continue is False):
                            self.enable()
                        return
                self._stop_requested = True
                return
            set_interrupt_handler(handle_interrupt)
            self._interrupt_continue(self) if self._interrupt_continue else None
            self.enable()
        def restore_interrupt_handler() -> None:  # noqa
            nonlocal self, previous_interrupt_handler
            set_interrupt_handler(previous_interrupt_handler)
        def set_interrupt_handler(interrupt_handler: Callable) -> Optional[Callable]:  # noqa
            nonlocal self
            if callable(interrupt_handler) and (threading.current_thread() == threading.main_thread()):
                return signal(SIGINT, interrupt_handler)
            return None
        previous_interrupt_handler = set_interrupt_handler(handle_interrupt)
        return namedtuple("interrupt_handler", ["restore"])(restore_interrupt_handler)

    def _define_tidy_output_hack(self) -> None:
        # Some minor tqdm output tidying-up which for annoying anomalies; tqdm forces
        # a colon (:) before the percentage, e.g. ":  25%|"; and while we're at it do
        # a little ASCII progress animation, requiring a special ([progress]) sentinal
        # string in the display string where the progress bar should actually go,
        # which we do in _format_description. Other minor things too; see below.
        sys_stdout_write = sys.stdout.write
        last_text = None ; last_captured_output_text = None  # noqa
        def tidy_stdout_write(text: str) -> None:  # noqa
            nonlocal self, sys_stdout_write, sentinel_internal, spina, spini, spinn
            nonlocal last_text, last_captured_output_text
            def replace_first(value: str, match: str, replacement: str) -> str:  # noqa
                return value[:i] + replacement + value[i + len(match):] if (i := value.find(match)) >= 0 else value
            def remove_extra_trailing_spaces(text: str) -> str:  # noqa
                while text.endswith("  "):
                    text = text[:-1]
                return text
            if (not text) or (last_text == text):
                return
            last_text = text
            if (self._disabled or self._done) and sentinel_internal in text:
                # Another hack to really disable output on interrupt; in this case we set
                # tqdm.disable to True, but output can still dribble out, so if the output
                # looks like it is from tqdm and we are disabled/done then do no output.
                return
            if sentinel_internal in text:
                spinc = spina[spini % spinn] if not ("100%|" in text) else "✓" ; spini += 1  # noqa
                text = replace_first(text, sentinel_internal, f" {spinc}")
                text = replace_first(text, "%|", "% ◀|")
                text = remove_extra_trailing_spaces(text)
                # Another oddity: for the rate sometimes tqdm intermittently prints
                # something like "1.54s/" rather than "1.54/s"; something to do with
                # the unit we gave, which is empty; idunno; just replace it here.
                text = replace_first(text, "s/ ", "/s ")
            if self._use_byte_size_for_rate and self._bar:
                rate = self._bar.n / (time.time() - self._started)
                text = text.replace("[rate]", f"{format_size(rate)}/s")
            sys_stdout_write(text)
            sys.stdout.flush()
            if self._captured_output_for_testing is not None:
                # For testing only we replace vacilliting values in the out like rate,
                # time elapsed, and ETA with static values; so that something like this:
                # > Working /  20% ◀|█████████▌  | 1/5 | 536.00/s | 00:01 | ETA: 00:02
                # becomes something more static like this after calling this function:
                # > Working |  20% ◀|### | 1/5 | 0.0/s | 00:00 | ETA: 00:00
                # This function obviously has intimate knowledge of the output; better here than in tests.
                def replace_time_dependent_values_with_static(text: str) -> str:
                    blocks = "\u2587|\u2588|\u2589|\u258a|\u258b|\u258c|\u258d|\u258e|\u258f"
                    if (n := find_nth_from_end(text, "|", 5)) >= 8:
                        pattern = re.compile(
                            rf"(\s*)(\d*%? ◀\|)(?:\s*{blocks}|#)*\s*(\|\s*\d+/\d+)?(\s*\|\s*)"
                            rf"(?:\d+\.?\d*|\?)(\/s\s*\|\s*)(?:\d+:\d+)?(\s*\|\s*ETA:\s*)(?:\d+:\d+|\?)?")
                        if match := pattern.match(text[n - 6:]):
                            if text[n - 8:n - 7] != "✓": text = set_nth(text, n - 8, "|")  # noqa
                            return (text[0:n - 6].replace("\r", "") +
                                    match.expand(rf"\g<1>\g<2>### \g<3>\g<4>0.0\g<5>00:00\g<6>00:00"))
                    return text
                if text != "\n":
                    captured_output_text = replace_time_dependent_values_with_static(text)
                    if captured_output_text != last_captured_output_text:
                        self._captured_output_for_testing.append(captured_output_text)
                        last_captured_output_text = captured_output_text
        def restore_stdout_write() -> None:  # noqa
            nonlocal sys_stdout_write
            if sys_stdout_write is not None:
                sys.stdout.write = sys_stdout_write
        sys.stdout.write = tidy_stdout_write
        spina = ["|", "/", "—", "◦", "\\"] ; spini = 0 ; spinn = len(spina)  # noqa
        sentinel = "[progress]" ; sentinel_internal = f"{sentinel}:"  # noqa
        return namedtuple("tidy_output_hack", ["restore", "sentinel"])(restore_stdout_write, sentinel)
